package org.codehaus.xfire.java.message;

import java.util.Calendar;
import java.util.Date;

import javax.xml.namespace.QName;
import javax.xml.stream.XMLStreamException;
import javax.xml.stream.XMLStreamReader;

import org.codehaus.xfire.XFireRuntimeException;
import org.codehaus.xfire.util.DateUtils;
import org.codehaus.xfire.util.DepthXMLStreamReader;

/**
 * Reads literal encoded messages.
 * 
 * @author <a href="mailto:dan@envoisolutions.com">Dan Diephouse</a>
 */
public class LiteralReader
    implements MessageReader
{
    private DepthXMLStreamReader root;
    private String value;
    private String localName;
    private QName name;
    private boolean hasCheckedChildren = false;
    private boolean hasChildren = false;
    private boolean hasFoundText = false;
    int depth;
    
    /**
     * Expects the XMLStreamReader in the START_DOCUMENT event.
     * @param root
     */
    public LiteralReader( DepthXMLStreamReader root )
    {
        this.root = root;
        this.localName = root.getLocalName();
        this.name = root.getName();
        depth = root.getDepth();
    }
    
    public LiteralReader( XMLStreamReader reader )
    {
        this.root = new DepthXMLStreamReader(reader);
        this.localName = root.getLocalName();
        this.name = root.getName();
        depth = root.getDepth();
    }

    /**
     * @see org.codehaus.xfire.java.message.MessageReader#getValue()
     */
    public String getValue()
    {
        if ( !hasFoundText )
            findValue();

        return value;
    }

    private void findValue()
    {
        while( !hasFoundText && checkHasMoreChildReaders() )
        {
        }
    }

    /**
     * @see org.codehaus.xfire.java.message.MessageReader#getValueAsInt()
     */
    public int getValueAsInt()
    {
        return Integer.parseInt( getValue() );
    }

    /**
     * @see org.codehaus.xfire.java.message.MessageReader#getValueAsDate()
     */
    public Date getValueAsDate()
    {
        if ( getValue() == null )
            return null;
        
        return DateUtils.parseDate( getValue() );
    }

    /**
     * @see org.codehaus.xfire.java.message.MessageReader#getValueAsDateTime()
     */
    public Date getValueAsDateTime()
    {
        if ( getValue() == null )
            return null;

        return DateUtils.parseDateTime( getValue() );
    }


    /**
	 * @see org.codehaus.xfire.java.message.MessageReader#getValueAsCalendar()
	 */
    public Calendar getValueAsCalendar()
    {
        if ( getValue() == null )
            return null;

        Calendar calendar = Calendar.getInstance();
        calendar.setTime( DateUtils.parseDateTime(getValue()) );
        return calendar;
    }

	/**
	 * @see org.codehaus.xfire.java.message.MessageReader#getValueAsLong()
	 */
	public long getValueAsLong()
	{
        return Long.parseLong( getValue() );
	}

	/**
	 * @see org.codehaus.xfire.java.message.MessageReader#getValueAsDouble()
	 */
	public double getValueAsDouble()
	{
		return Double.parseDouble( getValue() );
	}

	/**
	 * @see org.codehaus.xfire.java.message.MessageReader#getValueAsFloat()
	 */
	public float getValueAsFloat()
	{
		return Float.parseFloat( getValue() );
	}

	/**
	 * @see org.codehaus.xfire.java.message.MessageReader#getValueAsBoolean()
	 */
	public boolean getValueAsBoolean()
	{
		return Boolean.valueOf( getValue() ).booleanValue();
	}

    public boolean hasMoreChildReaders()
    {
        // Check to see if we checked before, 
        // so we don't mess up the stream position.
        if ( !hasCheckedChildren )
            checkHasMoreChildReaders();
        
        return hasChildren;
    }
    
    private boolean checkHasMoreChildReaders()
    {
        try
        {
            int event = root.getEventType();
            while ( true )
            {
                switch( event )
                {
                case XMLStreamReader.START_ELEMENT:
                    if ( root.getDepth() > depth )
                    {
                        hasCheckedChildren = true;
                        hasChildren = true;
                        return true;
                    }
                    break;
                case XMLStreamReader.END_ELEMENT:
                    if ( root.getDepth() <= depth )
                    {
                        hasCheckedChildren = true;
                        hasChildren = false;
                        root.next();
                        return false;
                    }
                    break;
                case XMLStreamReader.CHARACTERS:
                    value = root.getText();
                    hasFoundText = true;
                    break;
                case XMLStreamReader.END_DOCUMENT:
                    // We should never get here...
                    hasCheckedChildren = true;
                    hasChildren = false;
                    return false;
                default:
                    break;
                }
                event = root.next();
            }
        }
        catch (XMLStreamException e)
        {
            // TODO: I'm still not sure what should happen here.
            throw new XFireRuntimeException("Error parsing document.", e);
        }
    }

    public MessageReader getNextChildReader()
    {
        if ( !hasCheckedChildren )
            checkHasMoreChildReaders();
            
        if ( !hasChildren )
            return null;
        
        hasCheckedChildren = false;

        return new LiteralReader( root );
    }
    
    public QName getName()
    {
        return name;
    }

    public String getLocalName()
    {
        return localName;
    }
}
